"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fieldMaskPathsFromResolveInfo = void 0;
const graphql_1 = require("graphql");
/**
 * Create field mask paths from `GraphQLResolveInfo`.
 *
 * @param typename Name of GraphQL object type
 * @param info
 * @param opts
 * @returns field mask paths
 */
function fieldMaskPathsFromResolveInfo(typename, info, opts = {}) {
    return fieldMaskPaths(typename, info.fieldNodes, info.fragments, info.schema, opts);
}
exports.fieldMaskPathsFromResolveInfo = fieldMaskPathsFromResolveInfo;
function fieldMaskPaths(typename, nodes, fragments, schema, opts) {
    const pathSet = new Set();
    for (const node of nodes) {
        for (const path of extractFieldsFromGraphqlAst(typename, null, node, fragments, schema, opts)) {
            pathSet.add(path);
        }
    }
    return [...pathSet];
}
function extractFieldsFromGraphqlAst(typename, field, node, fragments, schema, opts) {
    if (!node.selectionSet) {
        return [];
    }
    const fields = [];
    for (const selection of node.selectionSet.selections) {
        switch (selection.kind) {
            case "Field": {
                if (selection.name.value === "__typename")
                    break;
                const type = getObjectType(typename, schema);
                const field = type.getFields()[selection.name.value];
                if (field == null) {
                    throw new Error(`${typename}.${selection.name.value} is not found`);
                }
                const fieldName = opts.getFieldName ? opts.getFieldName(field, type, schema) : field.name;
                if (fieldName == null) {
                    break;
                }
                if (selection.selectionSet) {
                    const childTypename = (0, graphql_1.getNamedType)(field.type).name;
                    const childFields = extractFieldsFromGraphqlAst(childTypename, field, selection, fragments, schema, opts);
                    fields.push(...childFields.map((field) => `${fieldName}.${field}`));
                }
                else {
                    fields.push(fieldName);
                }
                break;
            }
            case "FragmentSpread":
            case "InlineFragment": {
                let fragmentTypename;
                let node;
                switch (selection.kind) {
                    case "FragmentSpread": {
                        const fragmentName = selection.name.value;
                        const fragment = fragments[fragmentName];
                        if (fragment == null) {
                            throw new Error(`Fragment ${fragmentName} is not found`);
                        }
                        fragmentTypename = fragment.typeCondition.name.value;
                        node = fragment;
                        break;
                    }
                    case "InlineFragment": {
                        fragmentTypename = selection.typeCondition ? selection.typeCondition.name.value : typename;
                        node = selection;
                        break;
                    }
                }
                const currentNodeType = getType(typename, schema);
                if ((0, graphql_1.isAbstractType)(currentNodeType)) {
                    if (typename === fragmentTypename) {
                        const childFields = extractFieldsFromGraphqlAst(typename, field, node, fragments, schema, opts);
                        fields.push(...childFields);
                    }
                    else if (opts.getAbstractTypeFieldMaskPaths) {
                        if (field == null) {
                            throw new Error("field is expected to be present. please report issue.");
                        }
                        const paths = opts.getAbstractTypeFieldMaskPaths({ node, field, abstractType: currentNodeType, concreteType: getObjectType(fragmentTypename, schema) }, () => {
                            return extractFieldsFromGraphqlAst(fragmentTypename, field, node, fragments, schema, opts);
                        });
                        fields.push(...paths);
                    }
                    else {
                        // ignore union types if `getUnionFieldMaskPath` is not passed
                    }
                }
                else if (fragmentTypename !== typename) {
                    // no-op
                }
                else {
                    const childFields = extractFieldsFromGraphqlAst(typename, field, node, fragments, schema, opts);
                    fields.push(...childFields);
                }
                break;
            }
        }
    }
    return fields;
}
function getType(name, schema) {
    const foundType = schema.getType(name);
    if (foundType == null) {
        throw new Error(`${name} type is not found`);
    }
    return foundType;
}
function getObjectType(name, schema) {
    const foundType = getType(name, schema);
    if (!(foundType instanceof graphql_1.GraphQLObjectType)) {
        throw new Error(`${name} is ${foundType.astNode?.kind}, but want ObjectType`);
    }
    return foundType;
}
//# sourceMappingURL=fieldMaskPathsFromResolveInfo.js.map